@@ -8,9 +8,9 @@
 ;;             Rakotomandimby Mihamina <mihamina.rakotomandimby@rktmb.org>
 ;;             Bozhidar Batsov <bozhidar@batsov.dev>
 ;; URL: https://github.com/copilot-emacs/copilot.el
-;; Package-Requires: ((emacs "27.2") (editorconfig "0.8.2") (jsonrpc "1.0.14") (f "0.20.0") (track-changes "1.4"))
-;; Package-Version: 20251210.1016
-;; Package-Revision: 7ee4758bb748
+;; Package-Requires: ((emacs "27.2") (editorconfig "0.8.2") (jsonrpc "1.0.14") (compat "30") (track-changes "1.4"))
+;; Package-Version: 20260215.918
+;; Package-Revision: 189f0b4a5163
 ;; Keywords: convenience copilot
 
 ;; The MIT License (MIT)
@@ -40,12 +40,12 @@
 ;;; Code:
 
 (require 'cl-lib)
+(require 'compat)
 (require 'compile)
 (require 'json)
 (require 'jsonrpc)
 (require 'subr-x)
 
-(require 'f)
 (require 'editorconfig)
 (require 'track-changes)
 
@@ -166,7 +166,7 @@ find indentation offset."
   "The name of the package to install copilot server.")
 
 (defcustom copilot-install-dir (expand-file-name
-                                (locate-user-emacs-file (f-join ".cache" "copilot")))
+                                (locate-user-emacs-file (file-name-concat ".cache" "copilot")))
   "Directory in which the servers will be installed."
   :risky t
   :type 'directory
@@ -192,21 +192,16 @@ You may adjust this variable at your own risk."
 (define-obsolete-variable-alias 'copilot-version 'copilot-lsp-server-version "0.4.0")
 
 (defun copilot--lsp-settings-changed (symbol value)
-  "Restart the Copilot LSP due to SYMBOL changed to VALUE.
+  "Notify the Copilot LSP that SYMBOL changed to VALUE.
 
 This function will be called by the customization framework when the
 `copilot-lsp-settings' is changed.  When changed with `setq', then this function
 will not be called."
   (let ((was-bound (boundp symbol)))
     (set-default symbol value)
-    (when was-bound
-      ;; Notifying the agent with the new value does only work if we include the
-      ;; last value (as nil) as well. For example, having the value
-      ;; '(:github-enterprise (:uri "https://example2.ghe.com")) and setting it
-      ;; to nil would require to send the value '(:github-enterprise (:uri nil))
-      ;; to the server. Otherwise, the value is ignored, since sending nil is
-      ;; not enough.
-      (copilot--start-server))))
+    (when (and was-bound (copilot--connection-alivep))
+      (copilot--notify 'workspace/didChangeConfiguration
+                       `(:settings ,(copilot--effective-lsp-settings))))))
 
 (defcustom copilot-lsp-settings nil
   "Settings for the Copilot LSP server.
@@ -231,6 +226,16 @@ Exchange the URI with the correct URI of your organization."
   :group 'copilot
   :package-version '(copilot . "0.2"))
 
+(defcustom copilot-completion-model nil
+  "The completion model to use for Copilot suggestions.
+When nil, the server's default model is used.
+Use `M-x copilot-select-completion-model' to interactively choose
+from available models."
+  :type '(choice (const :tag "Default" nil)
+                 (string :tag "Model ID"))
+  :group 'copilot
+  :package-version '(copilot . "0.5"))
+
 (defvar-local copilot--overlay nil
   "Overlay for Copilot completion.")
 
@@ -261,6 +266,9 @@ Incremented after each change.")
 (defvar copilot--opened-buffers nil
   "List of buffers that have been opened in Copilot.")
 
+(defvar copilot--workspace-folders nil
+  "List of workspace folder URIs already reported to the server.")
+
 (defmacro copilot--dbind (pattern source &rest body)
   "Destructure SOURCE against plist PATTERN and eval BODY."
   (declare (indent 2))
@@ -311,9 +319,9 @@ Incremented after each change.")
   "Return the version number of currently installed `copilot-server-package-name'."
   (let ((possible-paths (list
                          (when (eq system-type 'windows-nt)
-                           (f-join copilot-install-dir "node_modules" copilot-server-package-name "package.json"))
-                         (f-join copilot-install-dir "lib" "node_modules" copilot-server-package-name "package.json")
-                         (f-join copilot-install-dir "lib64" "node_modules" copilot-server-package-name "package.json"))))
+                           (file-name-concat copilot-install-dir "node_modules" copilot-server-package-name "package.json"))
+                         (file-name-concat copilot-install-dir "lib" "node_modules" copilot-server-package-name "package.json")
+                         (file-name-concat copilot-install-dir "lib64" "node_modules" copilot-server-package-name "package.json"))))
     (seq-some
      (lambda (path)
        (when (and path (file-exists-p path))
@@ -333,10 +341,10 @@ Incremented after each change.")
    ((executable-find copilot-server-executable t))
    (t
     (let ((path (executable-find
-                 (f-join copilot-install-dir
-                       (cond ((eq system-type 'windows-nt) "")
-                             (t "bin"))
-                       copilot-server-executable)
+                 (file-name-concat copilot-install-dir
+                                   (cond ((eq system-type 'windows-nt) "")
+                                         (t "bin"))
+                                   copilot-server-executable)
                  t)))
       (unless (and path (file-exists-p path))
         (error "The package %s is not installed.  Unable to find %s"
@@ -475,6 +483,17 @@ SUCCESS-FN is the CALLBACK."
        ;; handle older jsonrpc versions
        (funcall make-fn :events-buffer-scrollback-size copilot-log-max)))))
 
+(defun copilot--effective-lsp-settings ()
+  "Return the effective LSP settings, including completion model."
+  (let ((settings (copy-sequence copilot-lsp-settings)))
+    (when copilot-completion-model
+      (let* ((github (or (plist-get settings :github) '()))
+             (copilot-section (or (plist-get github :copilot) '())))
+        (setq copilot-section (plist-put copilot-section :selectedCompletionModel copilot-completion-model))
+        (setq github (plist-put github :copilot copilot-section))
+        (setq settings (plist-put settings :github github))))
+    settings))
+
 (defun copilot--start-server ()
   "Start the copilot server process in local."
   (cond
@@ -487,23 +506,34 @@ SUCCESS-FN is the CALLBACK."
 You can change the installed version with `M-x copilot-reinstall-server` or remove this warning by changing the value of `copilot-lsp-server-version'."
               copilot-lsp-server-version installed-version)))
     (setq copilot--connection (copilot--make-connection))
+    (setq copilot--workspace-folders nil)
     (copilot--log 'info "Copilot server started.")
-    (copilot--request
-     'initialize
-     `(:processId
-       ,(emacs-pid)
-       :capabilities
-       (:workspace
-        (:workspaceFolders t))
-       :initializationOptions
-       (:editorInfo
-        (:name "Emacs" :version ,emacs-version)
-        :editorPluginInfo
-        (:name "copilot.el" :version ,(or (package-get-version) "unknown"))
-        ,@(when copilot-network-proxy
-            `(:networkProxy ,copilot-network-proxy)))))
+    (let* ((root (copilot--workspace-root))
+           (root-uri (when root (copilot--path-to-uri root)))
+           (folders (when root-uri
+                      (setq copilot--workspace-folders (list root-uri))
+                      (vector (list :uri root-uri :name (file-name-nondirectory (directory-file-name root)))))))
+      (copilot--request
+       'initialize
+       `(:processId
+         ,(emacs-pid)
+         ,@(when root-uri `(:rootUri ,root-uri))
+         :capabilities
+         (:workspace
+          (:workspaceFolders t)
+          :textDocument
+          (:inlineCompletion
+           (:dynamicRegistration :json-false)))
+         ,@(when folders `(:workspaceFolders ,folders))
+         :initializationOptions
+         (:editorInfo
+          (:name "Emacs" :version ,emacs-version)
+          :editorPluginInfo
+          (:name "copilot.el" :version ,(or (package-get-version) "unknown"))
+          ,@(when copilot-network-proxy
+              `(:networkProxy ,copilot-network-proxy))))))
     (copilot--notify 'initialized '())
-    (copilot--notify 'workspace/didChangeConfiguration `(:settings ,copilot-lsp-settings)))))
+    (copilot--notify 'workspace/didChangeConfiguration `(:settings ,(copilot--effective-lsp-settings))))))
 
 ;;
 ;; login / logout
@@ -514,7 +544,7 @@ You can change the installed version with `M-x copilot-reinstall-server` or remo
   (interactive)
   (copilot--dbind
       (status user ((:userCode user-code)) ((:verificationUri verification-uri)))
-      (copilot--request 'signInInitiate '(:dummy "signInInitiate"))
+      (copilot--request 'signInInitiate nil)
     (when (string-equal status "AlreadySignedIn")
       (user-error "Already signed in as %s" user))
     (if (display-graphic-p)
@@ -532,13 +562,13 @@ automatically, browse to %s." user-code verification-uri))
         (copilot--request 'signInConfirm (list :userCode user-code))
       (jsonrpc-error
        (user-error "Authentication failure: %s" (alist-get 'jsonrpc-error-message (cddr err)))))
-    (copilot--dbind (user) (copilot--request 'checkStatus '(:dummy "checkStatus"))
+    (copilot--dbind (user) (copilot--request 'checkStatus nil)
       (copilot--log 'info "Authenticated as GitHub user %s." user))))
 
 (defun copilot-logout ()
   "Logout from Copilot."
   (interactive)
-  (copilot--request 'signOut '(:dummy "signOut"))
+  (copilot--request 'signOut nil)
   (copilot--log 'warning "Logged out."))
 
 ;;
@@ -552,6 +582,7 @@ automatically, browse to %s." user-code verification-uri))
     (jsonrpc-shutdown copilot--connection 'kill)
     (setq copilot--connection nil))
   (setq copilot--opened-buffers nil)
+  (setq copilot--workspace-folders nil)
   ;; We are going to send a test request for the current buffer so we have to activate the mode
   ;; if it is not already activated.
   ;; If it the mode is already active, we have to make sure the current buffer is loaded in the
@@ -559,14 +590,10 @@ automatically, browse to %s." user-code verification-uri))
   (if copilot-mode
       (copilot--on-doc-focus (selected-window))
     (copilot-mode))
-  (copilot--async-request 'getCompletions
-                          `(:doc (:version 0
-                                           :source "\n"
-                                           :path ""
-                                           :uri ,(copilot--get-uri)
-                                           :relativePath ""
-                                           :languageId "text"
-                                           :position (:line 0 :character 0)))
+  (copilot--async-request 'textDocument/inlineCompletion
+                          (list :textDocument (list :uri (copilot--get-uri))
+                                :position '(:line 0 :character 0)
+                                :context '(:triggerKind 1))
                           :success-fn (lambda (_)
                                         (copilot--log 'info "Copilot OK."))
                           :error-fn (lambda (err)
@@ -574,6 +601,29 @@ automatically, browse to %s." user-code verification-uri))
                           :timeout-fn (lambda ()
                                         (copilot--log 'warning "Copilot server timeout."))))
 
+;;
+;; completion model selection
+;;
+
+(defun copilot-select-completion-model ()
+  "Interactively select a Copilot completion model."
+  (interactive)
+  (let* ((models (copilot--request 'copilot/models nil))
+         (completion-models
+          (seq-filter (lambda (m)
+                        (seq-contains-p (plist-get m :scopes) "completion"))
+                      models))
+         (choices (mapcar (lambda (m)
+                            (cons (format "%s (%s)" (plist-get m :modelName) (plist-get m :id))
+                                  (plist-get m :id)))
+                          completion-models))
+         (choice (completing-read "Completion model: " choices nil t))
+         (model-id (cdr (assoc choice choices))))
+    (setq copilot-completion-model model-id)
+    (copilot--notify 'workspace/didChangeConfiguration
+                     `(:settings ,(copilot--effective-lsp-settings)))
+    (copilot--log 'info "Completion model set to %s" model-id)))
+
 ;;
 ;; Auto completion
 ;;
@@ -639,30 +689,45 @@ automatically, browse to %s." user-code verification-uri))
           (display-warning '(copilot copilot-no-mode-indent)
                            "copilot--infer-indentation-offset found no mode-specific indentation offset.")
           (setq-local copilot--indent-warning-printed-p t))
-        standard-indent)))
+        tab-width)))
+
+(defun copilot--workspace-root ()
+  "Return the root directory of the current workspace, or nil."
+  (when buffer-file-name
+    (let ((root (or (and (fboundp 'project-current)
+                         (when-let* ((proj (project-current)))
+                           (project-root proj)))
+                    (and (fboundp 'projectile-project-root)
+                         (projectile-project-root))
+                    (and (fboundp 'vc-root-dir)
+                         (vc-root-dir)))))
+      (when root
+        (file-truename root)))))
 
 (defun copilot--get-relative-path ()
   "Get relative path to current buffer."
   (cond
    ((not buffer-file-name)
     "")
-   ((fboundp 'projectile-project-root)
-    (file-relative-name buffer-file-name (projectile-project-root)))
-   ((boundp 'vc-root-dir)
-    (file-relative-name buffer-file-name (vc-root-dir)))
    (t
-    (file-name-nondirectory buffer-file-name))))
+    (if-let* ((root (copilot--workspace-root)))
+        (file-relative-name buffer-file-name root)
+      (file-name-nondirectory buffer-file-name)))))
+
+(defun copilot--path-to-uri (path)
+  "Convert file PATH to a URI string."
+  (cond
+   ((and (eq system-type 'windows-nt)
+         (not (string-prefix-p "/" path)))
+    (concat "file:///" (url-encode-url path)))
+   (t
+    (concat "file://" (url-encode-url path)))))
 
 (defun copilot--get-uri ()
   "Get URI of current buffer."
-  (cond
-   ((not buffer-file-name)
-    (concat "file:///buffer/" (url-encode-url (buffer-name (current-buffer)))))
-   ((and (eq system-type 'windows-nt)
-         (not (string-prefix-p "/" buffer-file-name)))
-    (concat "file:///" (url-encode-url buffer-file-name)))
-   (t
-    (concat "file://" (url-encode-url buffer-file-name)))))
+  (if buffer-file-name
+      (copilot--path-to-uri buffer-file-name)
+    (concat "file:///buffer/" (url-encode-url (buffer-name (current-buffer))))))
 
 (defun copilot--get-source ()
   "Get source code from current buffer."
@@ -672,7 +737,8 @@ automatically, browse to %s." user-code verification-uri))
            (pmax (point-max))
            (pmin (point-min))
            (half-window (/ copilot-max-char 2)))
-      (when (and (>= copilot-max-char 0)
+      (when (and buffer-file-name
+                 (>= copilot-max-char 0)
                  (> pmax copilot-max-char))
         (let ((msg (format "%s size exceeds 'copilot-max-char' (%s), copilot completions may not work"
                            (current-buffer) copilot-max-char)))
@@ -720,13 +786,44 @@ automatically, browse to %s." user-code verification-uri))
   (or (copilot--get-minor-mode-language-id)
       (copilot--get-major-mode-language-id)))
 
+(defun copilot--utf16-offset ()
+  "Return the number of UTF-16 code units from line start to point.
+Characters above U+FFFF (e.g. emoji) count as 2 UTF-16 code units."
+  (let ((offset 0)
+        (p (line-beginning-position)))
+    (while (< p (point))
+      (let ((ch (char-after p)))
+        (setq offset (+ offset (if (>= ch #x10000) 2 1)))
+        (setq p (1+ p))))
+    offset))
+
+(defun copilot--utf16-strlen (str)
+  "Return the UTF-16 code-unit length of STR."
+  (let ((offset 0)
+        (i 0)
+        (len (length str)))
+    (while (< i len)
+      (let ((ch (aref str i)))
+        (setq offset (+ offset (if (>= ch #x10000) 2 1)))
+        (setq i (1+ i))))
+    offset))
+
+(defun copilot--goto-utf16-offset (utf16-offset)
+  "Move point forward by UTF16-OFFSET UTF-16 code units from line start.
+Point must be at line beginning before calling this."
+  (let ((remaining utf16-offset))
+    (while (and (> remaining 0) (not (eolp)))
+      (let ((ch (char-after)))
+        (setq remaining (- remaining (if (>= ch #x10000) 2 1))))
+      (forward-char 1))))
+
 (defun copilot--lsp-pos (&optional pos)
   "Return an LSP position plist for buffer POS.
-POS defaults to point."
+POS defaults to point.  Character offset is in UTF-16 code units."
   (save-excursion
     (when pos (goto-char pos))
     (list :line (- (line-number-at-pos) copilot--line-bias)
-          :character (- (point) (line-beginning-position)))))
+          :character (copilot--utf16-offset))))
 
 (defun copilot--generate-doc ()
   "Generate doc parameters for completion request."
@@ -744,37 +841,44 @@ POS defaults to point."
           :languageId (copilot--get-language-id)
           :position (copilot--lsp-pos))))
 
-(defun copilot--get-completion (callback)
-  "Get completion with CALLBACK."
-  (copilot--async-request 'getCompletions
-                          (list :doc (copilot--generate-doc))
-                          :success-fn callback))
+(defun copilot--inline-completion-params (trigger-kind)
+  "Build parameters for textDocument/inlineCompletion.
+TRIGGER-KIND is 1 for manual invocation, 2 for automatic."
+  (save-restriction
+    (widen)
+    (list :textDocument (list :uri (copilot--get-uri))
+          :position (copilot--lsp-pos)
+          :context (list :triggerKind trigger-kind)
+          :formattingOptions (list :tabSize (copilot--infer-indentation-offset)
+                                   :insertSpaces (if indent-tabs-mode :json-false t)))))
 
-(defun copilot--get-completions-cycling (callback)
-  "Get completion cycling options with CALLBACK."
-  (if copilot--completion-cache
-      (funcall callback copilot--completion-cache)
-    (copilot--async-request 'getCompletionsCycling
-                            (list :doc (copilot--generate-doc))
-                            :success-fn callback)))
+(defun copilot--normalize-completion-response (response)
+  "Normalize RESPONSE from textDocument/inlineCompletion to a list of items."
+  (cond
+   ((null response) nil)
+   ((vectorp response) (append response nil))
+   ((plist-get response :items)
+    (append (plist-get response :items) nil))
+   (t nil)))
+
+(defun copilot--get-completion (callback &optional trigger-kind)
+  "Get completion with CALLBACK.
+TRIGGER-KIND is 1 for invoked, 2 for automatic (default)."
+  (copilot--async-request 'textDocument/inlineCompletion
+                          (copilot--inline-completion-params (or trigger-kind 2))
+                          :success-fn callback))
 
 (defun copilot--cycle-completion (direction)
   "Cycle completion with DIRECTION."
-  (lambda (result)
-    (unless copilot--completion-cache
-      (setq copilot--completion-cache result))
-    (let ((completions (cl-remove-duplicates (plist-get result :completions)
-                                             :key (lambda (x) (plist-get x :text))
-                                             :test #'string-equal)))
-      (cond ((seq-empty-p completions)
-             (copilot--log 'warning "No completion is available."))
-            ((= (length completions) 1)
-             (copilot--log 'warning "Only one completion is available."))
-            (t (let ((idx (mod (+ copilot--completion-idx direction)
-                               (length completions))))
-                 (setq copilot--completion-idx idx)
-                 (let ((completion (elt completions idx)))
-                   (copilot--show-completion completion))))))))
+  (let* ((items copilot--completion-cache)
+         (len (length items)))
+    (cond ((or (null items) (zerop len))
+           (copilot--log 'warning "No completion is available."))
+          ((= len 1)
+           (copilot--log 'warning "Only one completion is available."))
+          (t
+           (setq copilot--completion-idx (mod (+ copilot--completion-idx direction) len))
+           (copilot--show-completion (nth copilot--completion-idx items))))))
 
 (defsubst copilot--overlay-visible ()
   "Return whether the `copilot--overlay' is available."
@@ -785,13 +889,13 @@ POS defaults to point."
   "Cycle to next completion."
   (interactive)
   (when (copilot--overlay-visible)
-    (copilot--get-completions-cycling (copilot--cycle-completion 1))))
+    (copilot--cycle-completion 1)))
 
 (defun copilot-previous-completion ()
   "Cycle to previous completion."
   (interactive)
   (when (copilot--overlay-visible)
-    (copilot--get-completions-cycling (copilot--cycle-completion -1))))
+    (copilot--cycle-completion -1)))
 
 (defvar copilot--panel-lang nil
   "Language of current panel solutions.")
@@ -854,7 +958,7 @@ Each request METHOD can have only one HANDLER."
                    "  :END:\n"
                    "#+BEGIN_SRC " copilot--panel-lang "\n"
                    completion-text "\n#+END_SRC\n\n")
-           (call-interactively #'mark-whole-buffer)
+           (goto-char (point-min))
            (org-sort-entries nil ?R nil nil "SCORE")))))))
 
 (copilot-on-notification
@@ -914,7 +1018,7 @@ To work around posn problems with after-string property.")
   "Keymap for Copilot completion overlay.")
 
 (defun copilot--posn-advice (&rest args)
-  "Remap posn if in copilot-mode with ARGS."
+  "Remap posn if in `copilot-mode' with ARGS."
   (when copilot-mode
     (let ((pos (or (car-safe args) (point))))
       (when (and copilot--real-posn
@@ -933,6 +1037,7 @@ To work around posn problems with after-string property.")
   "Create or get overlay for Copilot."
   (unless (overlayp copilot--overlay)
     (setq copilot--overlay (make-overlay 1 1 nil nil t))
+    (overlay-put copilot--overlay 'priority 100)
     (overlay-put
      copilot--overlay 'keymap-overlay (copilot--get-or-create-keymap-overlay)))
   copilot--overlay)
@@ -969,8 +1074,8 @@ To work around posn problems with after-string property.")
     (overlay-put ov 'completion completion)
     (overlay-put ov 'start (point))))
 
-(defun copilot--display-overlay-completion (completion uuid start end)
-  "Show COMPLETION with UUID between START and END.
+(defun copilot--display-overlay-completion (completion command full-insert-text start end)
+  "Show COMPLETION with COMMAND and FULL-INSERT-TEXT between START and END.
 
 `save-excursion' is not necessary since there is only one caller, and they are
 already saving an excursion.  This is also a private function."
@@ -980,18 +1085,17 @@ already saving an excursion.  This is also a private function."
     (let* ((ov (copilot--get-overlay)))
       (overlay-put ov 'tail-length (- (line-end-position) end))
       (copilot--set-overlay-text ov completion)
-      (overlay-put ov 'uuid uuid)
+      (overlay-put ov 'command command)
+      (overlay-put ov 'full-insert-text full-insert-text)
       (overlay-put ov 'completion-start start)
-      (copilot--async-request 'notifyShown (list :uuid uuid)))))
+      (when command
+        (copilot--notify 'textDocument/didShowCompletion
+                         (list :item (list :command command)))))))
 
-(defun copilot-clear-overlay (&optional is-accepted)
-  "Clear Copilot overlay.
-If IS-ACCEPTED is nil, notify rejected."
+(defun copilot-clear-overlay (&optional _is-accepted)
+  "Clear Copilot overlay."
   (interactive)
   (when (copilot--overlay-visible)
-    (unless is-accepted
-      (copilot--async-request 'notifyRejected
-                              (list :uuids `[,(overlay-get copilot--overlay 'uuid)])))
     (delete-overlay copilot--overlay)
     (delete-overlay copilot--keymap-overlay)
     (setq copilot--real-posn nil)))
@@ -1005,7 +1109,8 @@ provided."
     (let* ((completion (overlay-get copilot--overlay 'completion))
            (start (overlay-get copilot--overlay 'start))
            (end (copilot--overlay-end copilot--overlay))
-           (uuid (overlay-get copilot--overlay 'uuid))
+           (command (overlay-get copilot--overlay 'command))
+           (full-insert-text (overlay-get copilot--overlay 'full-insert-text))
            (t-completion (funcall (or transform-fn #'identity) completion))
            (completion-start (overlay-get copilot--overlay 'completion-start)))
       ;; If there is extra indentation before the point, delete it and shift the completion
@@ -1018,18 +1123,30 @@ provided."
         (setq start completion-start)
         (setq end (- end (- (point) completion-start)))
         (delete-region completion-start (point)))
-      (copilot--async-request 'notifyAccepted (list :uuid uuid))
-      (copilot-clear-overlay t)
-      (if (derived-mode-p 'vterm-mode)
-          (progn
-            (vterm-delete-region start end)
-            (vterm-insert t-completion))
-        (delete-region start end)
-        (insert t-completion))
-      ;; if it is a partial completion
-      (when (and (string-prefix-p t-completion completion)
-                 (not (string-equal t-completion completion)))
-        (copilot--set-overlay-text (copilot--get-overlay) (string-remove-prefix t-completion completion)))
+      (let ((is-partial (and (string-prefix-p t-completion completion)
+                             (not (string-equal t-completion completion)))))
+        (if is-partial
+            ;; Partial acceptance
+            (when command
+              (let* ((prefix-len (- (length full-insert-text) (length completion)))
+                     (accepted-length (+ prefix-len (length t-completion))))
+                (copilot--notify 'textDocument/didPartiallyAcceptCompletion
+                                 (list :item (list :command command)
+                                       :acceptedLength accepted-length))))
+          ;; Full acceptance
+          (when command
+            (copilot--async-request 'workspace/executeCommand
+                                    command)))
+        (copilot-clear-overlay t)
+        (if (derived-mode-p 'vterm-mode)
+            (progn
+              (vterm-delete-region start end)
+              (vterm-insert t-completion))
+          (delete-region start end)
+          (insert t-completion))
+        ;; if it is a partial completion, show remaining text
+        (when is-partial
+          (copilot--set-overlay-text (copilot--get-overlay) (string-remove-prefix t-completion completion))))
       t)))
 
 (defmacro copilot--define-accept-completion-by-action (func-name action)
@@ -1097,35 +1214,35 @@ Uppercase CHAR disables `case-fold-search', mirroring `zap-to-char'."
   "Show COMPLETION-DATA."
   (when (copilot--satisfy-display-predicates)
     (copilot--dbind
-        (text uuid ((:docVersion doc-version)) range)
+        (((:insertText insert-text)) command range)
         completion-data
-      (when (= doc-version copilot--doc-version)
-        (save-excursion
-          (save-restriction
-            (widen)
-            (let* ((p (point))
-                   (line (map-nested-elt range '(:start :line)))
-                   (start-char (map-nested-elt range '(:start :character)))
-                   (end-char (map-nested-elt range '(:end :character)))
-                   (goto-line! (lambda ()
-                                 (goto-char (point-min))
-                                 (forward-line (1- (+ line copilot--line-bias)))))
-                   (start (progn
-                            (funcall goto-line!)
-                            (forward-char start-char)
-                            (let* ((cur-line (buffer-substring-no-properties (point) (line-end-position)))
-                                   (common-prefix-len (length (copilot--string-common-prefix text cur-line))))
-                              (setq text (substring text common-prefix-len))
-                              (forward-char common-prefix-len)
-                              (point))))
-                   (end (progn
+      (save-excursion
+        (save-restriction
+          (widen)
+          (let* ((p (point))
+                 (full-insert-text insert-text)
+                 (line (map-nested-elt range '(:start :line)))
+                 (start-char (map-nested-elt range '(:start :character)))
+                 (end-char (map-nested-elt range '(:end :character)))
+                 (goto-line! (lambda ()
+                               (goto-char (point-min))
+                               (forward-line (1- (+ line copilot--line-bias)))))
+                 (start (progn
                           (funcall goto-line!)
-                          (forward-char end-char)
-                          (point)))
-                   (fixed-completion (copilot-balancer-fix-completion start end text)))
-              (goto-char p)
-              (pcase-let ((`(,start ,end ,balanced-text) fixed-completion))
-                (copilot--display-overlay-completion balanced-text uuid start end)))))))))
+                          (copilot--goto-utf16-offset start-char)
+                          (let* ((cur-line (buffer-substring-no-properties (point) (line-end-position)))
+                                 (common-prefix-len (length (copilot--string-common-prefix insert-text cur-line))))
+                            (setq insert-text (substring insert-text common-prefix-len))
+                            (forward-char common-prefix-len)
+                            (point))))
+                 (end (progn
+                        (funcall goto-line!)
+                        (copilot--goto-utf16-offset end-char)
+                        (point)))
+                 (fixed-completion (copilot-balancer-fix-completion start end insert-text)))
+            (goto-char p)
+            (pcase-let ((`(,start ,end ,balanced-text) fixed-completion))
+              (copilot--display-overlay-completion balanced-text command full-insert-text start end))))))))
 
 (defun copilot--on-doc-focus (window)
   "Notify that the document WINDOW has been focussed or opened."
@@ -1134,11 +1251,22 @@ Uppercase CHAR disables `case-fold-search', mirroring `zap-to-char'."
   ;; send a notification for the window gaining focus and only if the buffer has
   ;; copilot-mode enabled.
   (when (and copilot-mode (eq window (selected-window)))
+    ;; Detect new workspace roots and notify the server.
+    (when-let* ((root (copilot--workspace-root))
+                (root-uri (copilot--path-to-uri root)))
+      (unless (member root-uri copilot--workspace-folders)
+        (push root-uri copilot--workspace-folders)
+        (copilot--notify 'workspace/didChangeWorkspaceFolders
+                         (list :event
+                               (list :added (vector (list :uri root-uri
+                                                          :name (file-name-nondirectory
+                                                                 (directory-file-name root))))
+                                     :removed [])))))
     (if (seq-contains-p copilot--opened-buffers (current-buffer))
-        (copilot--notify ':textDocument/didFocus
+        (copilot--notify 'textDocument/didFocus
                          (list :textDocument (list :uri (copilot--get-uri))))
       (add-to-list 'copilot--opened-buffers (current-buffer))
-      (copilot--notify ':textDocument/didOpen
+      (copilot--notify 'textDocument/didOpen
                        (list :textDocument (list :uri (copilot--get-uri)
                                                  :languageId (copilot--get-language-id)
                                                  :version copilot--doc-version
@@ -1147,8 +1275,9 @@ Uppercase CHAR disables `case-fold-search', mirroring `zap-to-char'."
 (defun copilot--on-doc-close (&rest _args)
   "Notify that the document has been closed."
   (when (seq-contains-p copilot--opened-buffers (current-buffer))
-    (copilot--notify 'textDocument/didClose
-                     (list :textDocument (list :uri (copilot--get-uri))))
+    (when (copilot--connection-alivep)
+      (jsonrpc-notify copilot--connection 'textDocument/didClose
+                      (list :textDocument (list :uri (copilot--get-uri)))))
     (setq copilot--opened-buffers (delete (current-buffer) copilot--opened-buffers))))
 
 ;;;###autoload
@@ -1160,14 +1289,18 @@ Uppercase CHAR disables `case-fold-search', mirroring `zap-to-char'."
   (setq copilot--completion-cache nil)
   (setq copilot--completion-idx 0)
 
-  (let ((called-interactively (called-interactively-p 'interactive)))
+  (let ((called-interactively (called-interactively-p 'interactive))
+        (request-doc-version copilot--doc-version))
     (copilot--get-completion
-     (jsonrpc-lambda (&key completions &allow-other-keys)
-       (let ((completion (if (seq-empty-p completions) nil (seq-elt completions 0))))
-         (if completion
-             (copilot--show-completion completion)
-           (when called-interactively
-             (copilot--log 'warning "No completion is available."))))))))
+     (lambda (response)
+       (when (= request-doc-version copilot--doc-version)
+         (let ((items (copilot--normalize-completion-response response)))
+           (setq copilot--completion-cache items)
+           (if items
+               (copilot--show-completion (car items))
+             (when called-interactively
+               (copilot--log 'warning "No completion is available."))))))
+     (if called-interactively 1 2))))
 
 ;;
 ;; integration with track-changes
@@ -1188,17 +1321,17 @@ Uppercase CHAR disables `case-fold-search', mirroring `zap-to-char'."
                        (insert oldtext)
                        (goto-char (point-max))
                        (cons (1- (line-number-at-pos))
-                             (current-column))))
+                             (copilot--utf16-offset))))
            (num-newlines (car end-info))
            (end-char (cdr end-info)))
       (list :line (+ start-line num-newlines)
             :character (if (= num-newlines 0)
-                           (+ start-char (length oldtext))
+                           (+ start-char (copilot--utf16-strlen oldtext))
                          end-char)))))
 
 (defun copilot--track-changes-signal (id &optional _distance)
-  "Handle track changes signal for given tracker ID.
-Fetch changes and notify the language server."
+  "Handle `track-changes' signal for given tracker ID.
+Fetch the changes and notify the language server."
   (condition-case err
       (save-restriction
         (widen)
@@ -1336,7 +1469,7 @@ in `post-command-hook'."
   "Keymap for Copilot minor mode.
 Use this for custom bindings in `copilot-mode'.")
 
-(easy-menu-define copilot-mode-menu copilot-mode-map "Copilot Menu"
+(easy-menu-define copilot-mode-menu copilot-mode-map "Copilot menu."
   '("Copilot"
     ["Accept Completion" copilot-accept-completion]
     ["Accept Completion by Word" copilot-accept-completion-by-word]
@@ -1394,8 +1527,12 @@ Use this for custom bindings in `copilot-mode'.")
     (copilot--mode-teardown)))
 
 (defun copilot-turn-on-unless-buffer-read-only ()
-  "Turn on `copilot-mode' if the buffer is writable."
-  (unless buffer-read-only
+  "Turn on `copilot-mode' if the buffer is writable and not internal.
+Skip read-only buffers, minibuffers, and hidden internal buffers
+whose names start with a space."
+  (unless (or buffer-read-only
+              (minibufferp)
+              (string-prefix-p " " (buffer-name)))
     (copilot-mode 1)))
 
 ;;;###autoload
